from swarmform import SwarmPad
from fireworks import Firework, ScriptTask, FWorker
from fireworks.core.rocket_launcher import launch_rocket

if __name__ == "__main__":
	# set up the SwarmPad and reset it
	swarmpad = SwarmPad()
	swarmpad.reset('', require_password=False)

	# create the Firework consisting of a custom "Addition" task
	firework = Firework(ScriptTask.from_str('echo "hello"'))

	# store workflow
	swarmpad.add_sf(firework)

	# Retrieve SwarmFlow from the SwarmPad
	sf = swarmpad.get_sf_by_id(1)

	sf = swarmpad.get_sf_by_name('Unnamed FW')

	# Run the swarmFlow
	launch_rocket(swarmpad, FWorker())



from swarmform import SwarmPad, WorkflowGenerator


def main():
	# set up the LaunchPad and reset it
	swarmpad = SwarmPad()
	swarmpad.reset('', require_password=False)
	workflow_yaml = "/home/kalana/fyp-new/SwarmForm/swarmform/util/workflows/yaml/custom-jobs-25.yaml"
	dax_swarmflow = WorkflowGenerator.generate_workflow(workflow_yaml)
	swarmpad.add_sf(dax_swarmflow)


if __name__ == "__main__":
	main()



import datetime
import os
import yaml
import xml.etree.ElementTree as ET

from fireworks import Firework, ScriptTask
from swarmform.core.swarmwork import SwarmFlow


class WorkflowGenerator():

    @classmethod
    # Read input from DAX file
    def read_input_dax(cls, filename):
        try:
            with open(filename, 'r') as f:
                input_xml = ET.parse(filename)
                return input_xml
        except IOError:
            print("Error while opening the DAX file")

    @classmethod
    def parse_dax(cls, xml_tree):
        root = xml_tree.getroot()
        workflow_dict = {}
        id_map = {}
        fw_id = 1

        job_count = root.attrib['jobCount']
        namespace = root[0].attrib['namespace']
        swarmflow_name = namespace + "_" + job_count
        for child in root:
            if child.tag.endswith('job'):
                # Set execution time as the first element of the firework
                # Set cores to zero as it is considered only in parallel task clustering
                # Set empty list to add children
                firework = [round((float(child.attrib['runtime'])/float(10)), 3), 0]
                workflow_dict.update({fw_id: firework})

                # Create a mapping between dax job id and firework id
                id_map.update({child.attrib['id']: fw_id})
                fw_id += 1

        dependency_dict = cls.get_parent_child_relationships(id_map, root)
        updated_dependency_dict = cls.replace_job_id_with_fw_id(id_map, dependency_dict)
        for fw_id in workflow_dict:
            fw = workflow_dict[fw_id]
            fw.append(updated_dependency_dict[fw_id])

        return workflow_dict, swarmflow_name

    @classmethod
    # When the DAX is given, output a dict with children of each parent
    # { parent_id : [child_id] }
    def get_parent_child_relationships(cls, id_map, dax):
        job_ids = id_map.keys()
        dependency_dict = {}
        for job in job_ids:
            # Get the child tag in the dict and set it as the parent
            parent = job
            children = []
            # Traverse the DAX to find children with parent tags equal to parent
            # If found, set the child tag(one level up in the xml tree) as a child of the considered parent
            for child in dax:
                if child.tag.endswith('child'):
                    for parent_id in child:
                        if parent_id.attrib['ref'] == parent:
                            children.append(child.attrib['ref'])
            dependency_dict.update({parent: children})
        return dependency_dict

    @classmethod
    def replace_job_id_with_fw_id(cls, id_map,dependency_dict):

        # Enumerate dependency_dict and replace its children_ids with the fw_ids of the children
        for parent_id in id_map:
            child_id_list = dependency_dict[parent_id]
            for index, child_id in enumerate(child_id_list):
                if child_id in id_map:
                    child_id_list[index] = id_map[child_id]

            # Enumerate dependency_dict and replace its parent ids with their fw_ids
            if parent_id in id_map:
                dependency_dict[id_map[parent_id]] = dependency_dict.pop(parent_id)

        return dependency_dict

    @classmethod
    # Read input from YAML file
    def read_input_yaml(cls, filename):
        try:
            with open(filename, 'r') as f:
                input_yaml = yaml.load(f, Loader=yaml.FullLoader)
            jobs = input_yaml['fireworks']
            swarmflow_name = input_yaml['swarmflow_name']
            return jobs, swarmflow_name
        except IOError:
            print("Error while opening the DAX file")

    @classmethod
    # Generate script for each shell script
    def gen_script(cls, job_id, job_exec_time):
        script = """
                    start=$(date +"%T.%3N");
                    echo "##########task {} start time ${{start}}";
                    sleep {};
                    end=$(date +"%T.%3N");
                    echo "##########task {} end time ${{end}}";
                    exit;
                    """
        script = script.format(job_id, job_exec_time, job_id)
        return script

    @classmethod
    # Create directory for the swarmflow
    def create_directory(cls, swarmflow_name):
        dir_name = swarmflow_name + "-" + str(int(datetime.datetime.now().timestamp()))
        os.mkdir(dir_name)
        return dir_name

    @classmethod
    # Create firework for each shell script
    def create_firework(cls, filename):
        cur_dir = os.getcwd()
        task_path = cur_dir + "/" + filename
        command = "sh " + task_path
        task = ScriptTask.from_str(command)
        firework = Firework(task)
        return firework

    @classmethod
    # Create shell script file for each firework
    def create_scripts(cls, dir_name, jobs):
        fws = []

        # eg: filename = task1.sh
        for job_id in jobs:
            filename = dir_name + "/task" + str(job_id) + ".sh"
            job_exec_time = jobs[job_id][0]
            script = cls.gen_script(job_id, job_exec_time)

            with open(filename, 'w+') as f:
                f.write(script)

        # Create firework for each job given
            firework = cls.create_firework(filename)
            fws.append(firework)

        return fws

    @classmethod
    # Map the firework ids to its list positions
    def map_list_positions(cls, child_id):
        index = child_id - 1
        return index

    @classmethod
    # Create metadata for each firework
    def create_metadata(cls, jobs, fireworks):

        # output format { 'costs' : {fw_id : {'exec_time' : x , 'cores' : y }}}
        metadata = {'costs': {}}
        for job_id in jobs:
            # 1st element in jobs dictionary represent execution time of the firework
            # 2nd element in jobs dictionary represent cores required for the firework
            exec_time = jobs[job_id][0]
            cores = jobs[job_id][1]

            # Get the position of firework in the fireworks list
            parent_fw_position = job_id - 1
            metadata_dict = {'exec_time': exec_time, 'cores': cores}
            fw_id = fireworks[parent_fw_position].fw_id

            # Create metadata dictionary for each firework and add it to metadata dictionary
            metadata['costs'].update({fw_id: metadata_dict})

        return metadata

    @classmethod
    # Create the dependency dictionary using the given parent-child relationships
    def create_dependencies(cls, jobs, fireworks):
        dependencies = {}

        for job_id in jobs:
            # 3rd element in jobs dictionary represent children of the firework
            children = jobs[job_id][2]

            # Get the position of firework in the fireworks list, relevant to the considered firework
            parent_fw_position = job_id - 1
            if children is not None and children != []:
                children_fw_positions = map(cls.map_list_positions, children)

                # Create a children fireworks list for each firework and add it to dependencies dictionary
                child_list = []
                for position in children_fw_positions:
                    child_list.append(fireworks[position])
            else:
                child_list = []
            parent_fw = fireworks[parent_fw_position]

            dependencies.update({parent_fw: child_list})

        return dependencies

    @classmethod
    # Save the swarmflow to a YAML file
    def dump_swarmflow(cls, swarmflow, dir_name, swarmflow_name):
        swarmflow_output_filename = dir_name + "/" + swarmflow_name + ".yaml"
        swarmflow.to_file(swarmflow_output_filename, "yaml")
        return None

    @classmethod
    def generate_workflow(cls, input_file):

        if input_file.endswith('yaml') or input_file.endswith('yml'):
            jobs, swarmflow_name = cls.read_input_yaml(input_file)
        elif input_file.endswith('xml'):
            xml_tree = cls.read_input_dax(input_file)
            jobs, swarmflow_name = cls.parse_dax(xml_tree)
        else:
            raise IOError('Input file format not recognized. Only YAML and DAX formats are supported')
        dir_name = cls.create_directory(swarmflow_name)
        fireworks = cls.create_scripts(dir_name, jobs)
        dependencies = cls.create_dependencies(jobs, fireworks)
        metadata = cls.create_metadata(jobs, fireworks)
        swarmflow = SwarmFlow(fireworks=fireworks, links_dict=dependencies, metadata=metadata, name=swarmflow_name)
        cls.dump_swarmflow(swarmflow, dir_name, swarmflow_name)
        return swarmflow




class Node:

    def __init__(self, fw_id, level, fw_info, assigned=False, parents=None, children=None):

        """
        Args:
            fw_id (int): an identification number for this Firework.
            level (int): depth of the node from the entry node.
            fw_info (dict): contains info like execution time in seconds, required number of cores of a node(job).
                            eg: fw_info = {'exec_time': 10, 'cores': 5}
            assigned (boolean): indicates whether this node is assigned to a cluster or not
            parents (Node or [Node]): list of parent nodes this node depends on.
            children (Node or [Node]): list of children nodes depend on this node.
        """

        self._fw_id = fw_id
        self._level = level
        self._fw_info = fw_info
        self._is_assigned = assigned
        self._parents = parents
        self._children = children
        self._cluster_info = {fw_id: fw_info}
        self._sequential_ids = []
        self._parallel_ids = {}
        self._cluster_space = []

    def set_cluster_space(self, space):
        self._cluster_space = space

    def get_cluster_space(self):
        return self._cluster_space

    def set_sequential_ids(self, id_list):
        self._sequential_ids = id_list

    def get_fw_ids_to_cluster_sequentially(self):
        return self._sequential_ids

    def set_parallel_ids(self, id_list, key):
        self._parallel_ids[key] = id_list

    def get_fw_ids_to_cluster_parallely(self):
        return self._parallel_ids

    def get_fw_id(self):
        return self._fw_id

    def get_level(self):
        return self._level

    def get_exec_time(self):
        if self._fw_info:
            return self._fw_info['exec_time']
        else:
            return None

    def get_num_cores(self):
        if self._fw_info:
            return self._fw_info['cores']
        else:
            return None

    def get_is_assigned(self):
        return self._is_assigned

    def get_parents(self):
        return self._parents

    def get_children(self):
        return self._children

    def get_cluster_info(self):
        return self._cluster_info

    def set_cluster_info(self, c_info):
        """
        Args:
            c_info (dict): a dictionary of cluster information containing {fw_id: fw_info}
        """
        self._cluster_info = c_info

    def get_fw_ids_to_cluster(self):
        fw_ids_to_cluster = self._cluster_info.keys()
        return list(fw_ids_to_cluster)

    def add_parent(self, parent):

        if not isinstance(self._parents, list):
            self._parents = []

        self._parents.append(parent)

    def add_child(self, child):

        if not isinstance(self._children, list):
            self._children = []

        self._children.append(child)

    def remove_parent(self, parent_id):

        for parent in self._parents:
            if parent.get_fw_id() == parent_id:
                self._parents.remove(parent)
                break

    def remove_child(self, child_id):

        for child in self._children:
            if child.get_fw_id() == child_id:
                self._children.remove(child)
                break

    def set_fw_info(self, exec_time, cores):
        self._fw_info = {'exec_time': exec_time, 'cores': cores}


class DAG:

    def __init__(self, sf):

        """
        Args:
            sf(SwarmFlow): SwarmFlow object
        """

        self._dag_id = sf.sf_id
        self._dag_name = sf.name or "Unnamed-SF"
        self._nodes = {}  # dictionary in format of {fw_id: Node}
        parents_dict = {}

        fireworks = sf.fws
        self._links = sf.links  # dictionary in the format of {parent_id:[child_ids]}
        metadata = sf.metadata  # dictionary in the format of {fw_id: [exec_time, cores]}
        self._costs = metadata['costs'] if 'costs' in metadata else {}

        # creating Nodes and adding to the _nodes dictionary
        self._height = 0
        for fw in fireworks:
            fw_id = fw.fw_id
            level = self.find_node_level(fw_id)
            fw_info = self._costs[str(fw_id)] if str(fw_id) in self._costs else {}
            node = Node(fw_id=fw_id, level=level, fw_info=fw_info)
            if fw_id in self._nodes or fw_id in parents_dict:
                raise ValueError('FW ids must be unique!')
            self._nodes[fw_id] = node
            if level > self._height:
                self._height = level

        # setting parents and children
        for parent_id in self._links:
            parent_node = self._nodes[parent_id]
            children = self._links[parent_id]
            for child_id in children:
                child_node = self._nodes[child_id]
                parent_node.add_child(child_node)
                child_node.add_parent(parent_node)

    def get_dag_id(self):
        return self._dag_id

    def get_dag_name(self):
        return self._dag_name

    def get_height(self):
        return self._height

    def get_nodes(self):
        return self._nodes

    def get_costs(self):
        return self._costs

    def get_parent_child_relationships(self):
        return self._links

    def find_all_paths(self, start, end, path=[]):
        """
        Args:
            start (int): starting node id
            end (int): ending node id
            path (list): list of node ids from starting node to ending node

        Returns:
            paths (list[list]): 2D list containing all the paths to ending node
        """
        graph = self._links
        path = path + [start]
        if start == end:
            return [path]
        if start not in graph.keys():
            return []
        paths = []
        for node in graph[start]:
            if node not in path:
                new_paths = self.find_all_paths(node, end, path)
                for new_path in new_paths:
                    paths.append(new_path)

        return paths

    def all_paths_from_roots(self, node_id):
        """
        Args:
            node_id (int): id of the node which needs to be traversed

        Returns:
            paths (list[list]): 2D list containing all the paths from all root nodes to given node
        """
        graph = self._links
        paths = []
        for key in graph.keys():
            if all(key not in val for val in graph.values()):
                paths += self.find_all_paths(key, node_id)

        return paths

    def find_node_level(self, node_id):
        """
        Args:
            node_id (int): node id of the node of level required

        Returns:
            level (int)
        """
        paths = self.all_paths_from_roots(node_id)

        return max([len(x) for x in paths])

    def add_node(self, fw_id, node):
        """
        Args:
            fw_id (int): firework id of the node
            node (Node): node object needs to be added
        """
        if fw_id in self._nodes:
            raise ValueError('FW ids must be unique!')
        self._nodes[fw_id] = node

    def delete_node(self, fw_id):
        """
        Args:
            fw_id (int): firework id of the node to be deleted
        """
        if fw_id not in self._nodes:
            raise KeyError('FW id not exists')
        del self._nodes[fw_id]

    def update_links(self):
        """
            Reset the links attribute and update with new parent-children relationships.
            This method should be called after overriding the DAG object.
        """
        links = {}
        for fw_id in self._nodes:
            node = self._nodes[fw_id]
            children = node.get_children()
            if children:
                links[node.get_fw_id()] = []
                for child in children:
                    links[node.get_fw_id()].append(child.get_fw_id())
        self._links = links

    def update_height(self):
        """
        Update the height attribute with updated links
        This method should be called after overriding the DAG object if the height is changed
        """
        self.update_links()
        height = 0
        for fw_id in self._nodes:
            level = self.find_node_level(fw_id)
            if level > height:
                height = level

        self._height = height



from fireworks import Firework, ScriptTask
from swarmform import ParallelTask
from swarmform.core.swarm_dag import DAG
from swarmform.core.clustering_algo.wpa_clustering import wpa_clustering, cluster_vertically
from swarmform.core.swarmwork import SwarmFlow


def combine_fws_sequentially(swarmpad, fw_ids, parallely_clustered_fws, parallely_clustered_fw_ids):
    """
    Combine a set of fireworks into a single firework

    Args:
        swarmpad (SwarmPad)
        fw_ids (list): id of the fireworks to be combined sequentially
        parallely_clustered_fws(list): list of fireworks which are clustered parallely,but not added to the SwarmPad
        parallely_clustered_fw_ids(dict): dictionary of {cluster id: firework id } of parallely clustered fireworks

    Returns:
        combinedFW (Firework)
    """

    fws = parallely_clustered_fws

    parallel_fw_ids = parallely_clustered_fw_ids.values()

    firetask = []

    # Get the firetasks of parallel fireworks and add to firetask list
    for parallel_fw in fws:
        for task in parallel_fw.tasks:
            firetask.append(task)

    # Get firework from swarmpad if it is not available in parallely_clustered_fws list
    for fw_id in fw_ids:
        if fw_id not in parallel_fw_ids:
            firetask_list = swarmpad.get_fw_by_id(fw_id).spec['_tasks']
            num_firetasks = len(firetask_list)
            # Check whether a firework has no firetasks
            if num_firetasks == 0:
                raise ValueError('No Firetasks available in the Firework')
            # Check whether the firework has multiple firetasks - Not checking firetasks recursively at the moment
            if num_firetasks > 1:
                # If multiple firetasks are available, add each task to firetask list
                for firetask_index in range(0, num_firetasks):
                    firetask.append(firetask_list[firetask_index])
            # If only a single firetask is there, get the first firetask and add it to the combined firetask
            else:
                firetask.append(firetask_list[0])

    # Create a firework from the combined firetasks
    combined_fw = Firework(firetask)
    swarmpad.m_logger.info('Sequentially clustered {} Fireworks to firework_id {}'.format(fw_ids, combined_fw.fw_id))
    return combined_fw


# TODO: Resolve the following assumptions
'''
* Only scriptTasks are given
* Command in scriptTasks are static
* fws to combine parallely have only a single firetask
'''


def combine_fws_parallely(swarmpad, fw_ids):
    """
    Combine a set of firetasks into a single firetask which runs all the given tasks parallely

    Args:
        swarmpad (SwarmPad)
        fw_ids (list): id of the fireworks to be combined

    Returns:
        combined_firework (FireWork): Parallely combined FireWork object
    """

    # Get each task in each firework and append to firetask list in the order of traversal
    firetasks_to_combine = []
    for fw_id in fw_ids:
        firetask_in_fw = swarmpad.get_fw_by_id(fw_id).spec['_tasks'][0]
        if isinstance(firetask_in_fw, ScriptTask):
            firetasks_to_combine.append(firetask_in_fw)
        else:
            raise ValueError('Spec of Firework with id {} does not contain an object of type ScriptTask '.format(fw_id))

    combined_firetask = ParallelTask.from_firetasks(firetasks_to_combine)
    combined_firework = Firework(combined_firetask)
    swarmpad.m_logger.info('Parallely Clustered {} to firework_id {}'.format(fw_ids, combined_firework.fw_id))

    return combined_firework


def update_parent_child_relationships(links_dict, old_id, new_id):
    """
    Update the parent-child relationships after clustering a firework
    by replacing all the instances of old_id with new_id
    Args:
        links_dict (list): Existing parent-child relationship list
        old_id (int): Existing id of the firework
        new_id (int): New id of the firework

    Returns:
        links_dict (list): Updated parent-child relationship list
    """

    # Enumerate child ids and replace it with the new id
    for parent_id in links_dict:
        child_id_list = links_dict[parent_id]
        for index, child_id in enumerate(child_id_list):
            if child_id == old_id:
                child_id_list[index] = new_id
                break

    # Enumerate parent ids and replace it with the new id
    if old_id in links_dict:
        links_dict[new_id] = links_dict.pop(old_id)
    return links_dict


def cluster_sf(swarmpad, sf_id):

    """
    Pull the swarmflow from given sf_id and create a clustered swarmflow

    Args:
        swarmpad (SwarmPad)
        sf_id (int): id of the swarmflow to pull

    Returns:
        Clustered_swarmflow (SwarmFlow)
    """
    # Retrieve the relevant swarmflow from the swarmpad
    sf = swarmpad.get_sf_by_id(sf_id)
    sf_dag = DAG(sf)
    vertically_clustered_dag = cluster_vertically(sf_dag)
    # Cluster the swarmflow DAG
    clustered_sf_dag = wpa_clustering(vertically_clustered_dag)
    # Get parent-child relationships of the clustered dag {cluster_id : [fw_ids] }
    # eg: links {17: [18, 19, 21, 20], 18: [23], 19: [23], 21: [23], 20: [23]}
    links_dict = clustered_sf_dag.get_parent_child_relationships()
    # Get nodes of the clustered dag { cluster_id: Node }
    nodes = clustered_sf_dag.get_nodes()
    clustered_fws = []

    for key in nodes:
        # Dictionary of parallel clusters
        fw_ids_to_cluster_parallely = nodes[key].get_fw_ids_to_cluster_parallely()
        # List of sequential fireworks
        fw_ids_to_cluster_sequentially = nodes[key].get_fw_ids_to_cluster_sequentially()

        # List for storing parallely clustered fireworks
        parallely_clustered_fws = []
        # Dict for storing new firework id of the clustered firework with respect to its cluster id
        parallely_clustered_fw_ids = {}

        if len(fw_ids_to_cluster_parallely) != 0:
            for cluster_id in fw_ids_to_cluster_parallely:
                parallely_clustered_fw = combine_fws_parallely(swarmpad, fw_ids_to_cluster_parallely[cluster_id])
                parallely_clustered_fws.append(parallely_clustered_fw)
                parallely_clustered_fw_ids.update({cluster_id: parallely_clustered_fw.fw_id})

        # Replace the cluster id in sequential fws to cluster with its new fw id
        fw_ids_to_cluster_sequentially = [parallely_clustered_fw_ids.get(n, n) for n in
                                          fw_ids_to_cluster_sequentially]

        # If multiple fireworks are available, cluster them and to clustered_fws
        if len(fw_ids_to_cluster_sequentially) > 1:
            combined_fw = combine_fws_sequentially(swarmpad, fw_ids_to_cluster_sequentially, parallely_clustered_fws,
                                                   parallely_clustered_fw_ids)
            for fw_id in fw_ids_to_cluster_sequentially:
                links_dict = update_parent_child_relationships(links_dict, fw_id, combined_fw.fw_id)

        # If only a single firework is available, add it directly to clustered_fws
        elif len(fw_ids_to_cluster_sequentially) == 0 or len(fw_ids_to_cluster_sequentially) == 1:
            combined_fw = swarmpad.get_fw_by_id(nodes[key].get_fw_id())
        else:
            raise ValueError(
                "Issue with the firework ids to cluster: {}".format(fw_ids_to_cluster_sequentially))
        clustered_fws.append(combined_fw)

    clustered_swarmflow = SwarmFlow(fireworks=clustered_fws, links_dict=links_dict)
    return clustered_swarmflow



from fireworks import Workflow, Firework


class SwarmFlow(Workflow):

    def __init__(self, fireworks, links_dict=None, name=None, metadata=None, created_on=None, updated_on=None,
                 fw_states=None, sf_id=None):
        """
        Args:
            fireworks ([Firework]): all FireWorks in this SwarmFlow.
            links_dict (dict): links between the FWs as (parent_id):[(child_id1, child_id2)]
            name (str): name of the SwarmFlow.
            metadata (dict): metadata for this SwarmFlow.
            created_on (datetime): time of creation
            updated_on (datetime): time of update
            fw_states (dict): leave this alone unless you are purposefully creating a Lazy-style WF
            sf_id (int): SwarmFlow id
        """

        super().__init__(fireworks, links_dict, name, metadata, created_on, updated_on, fw_states)

        if sf_id is not None:
            self.sf_id = sf_id

        self.fw_costs = self.metadata['costs'] if 'costs' in self.metadata else {}

    def _reassign_ids(self, old_new):
        """
        Internal method to reassign Firework ids, e.g. due to database insertion.

        Args:
                old_new (dict)
        """

        super()._reassign_ids(old_new)

        # update fw costs
        new_fw_costs = {}
        for (fwid, cost) in self.fw_costs.items():
            new_fw_costs[str(old_new.get(int(fwid), int(fwid)))] = cost
        self.fw_costs = new_fw_costs
        self.metadata['costs'] = self.fw_costs

    def to_db_dict(self):
        m_dict = super().to_db_dict()
        m_dict['sf_id'] = self.sf_id
        return m_dict

    @classmethod
    def from_dict(cls, m_dict):
        """
        Return Workflow from its dict representation.

        Args:
            m_dict (dict): either a Workflow dict or a Firework dict

        Returns:
            SwarmFlow
        """

        if 'fws' in m_dict:
            created_on = m_dict.get('created_on')
            updated_on = m_dict.get('updated_on')
            return SwarmFlow([Firework.from_dict(f) for f in m_dict['fws']],
                            Workflow.Links.from_dict(m_dict['links']), m_dict.get('name'),
                            m_dict['metadata'], created_on, updated_on)
        else:
            return SwarmFlow.from_Firework(Firework.from_dict(m_dict))

    @classmethod
    def from_Firework(cls, fw, name=None, metadata=None, sf_id=None):
        """
        Return SwarmFlow from the given Firework.

        Args:
            fw (Firework)
            name (str): New SwarmFlow's name. if not provided, the firework name is used
            sf_id (int): Id of the SwarmFlow
            metadata (dict): New SwarmFlow's metadata.

        Returns:
            SwarmFlow
        """

        name = name if name else fw.name
        return SwarmFlow([fw], name=name, metadata=metadata, created_on=fw.created_on,
                         updated_on=fw.updated_on, sf_id=sf_id)



from swarmform import SwarmPad
from swarmform.core.swarmwork import SwarmFlow
from swarmform.core.cluster import cluster_sf

if __name__ == "__main__":
	# set up the SwarmPad and reset it
	swarmpad = SwarmPad()
	swarmpad.reset('', require_password=False)

	filename = '/swarmform/custom-25-jobs-correct-1587484561/custom-25-jobs-correct.yaml'
	# create the Firework consisting of a custom "Addition" task
	unclustered_sf = SwarmFlow.from_file(filename)

	# store workflow
	swarmpad.add_sf(unclustered_sf)

	sf = swarmpad.get_sf_by_id(unclustered_sf.sf_id)

	# Cluster the SwarmFlow
	clustered_sf = cluster_sf(swarmpad, unclustered_sf.sf_id)

	# Archive unclustered SwarmFlow
	unclustered_sf_fw_id = unclustered_sf.fws[0].fw_id

	swarmpad.archive_wf(unclustered_sf_fw_id)

	swarmpad.add_sf(clustered_sf)



# coding: utf-8

from __future__ import unicode_literals
from fireworks import ScriptTask, explicit_serialize


@explicit_serialize
class ParallelTask(ScriptTask):
	""" Combines multiple user defined ScriptTasks parallely"""
	required_params = ['script']

	@classmethod
	def from_firetasks(cls, firetasks, parameters=None):
		firetasks = firetasks if isinstance(firetasks, (list, tuple)) else [firetasks]
		parameters = parameters if parameters else {}
		script = ''
		for task in firetasks:
			script += task['script'][0] + ' & '
		script += 'wait'
		parameters['script'] = [script]
		parameters['use_shell'] = True
		return cls(parameters)



from swarmform.core.swarm_dag import Node, DAG
from random import randrange

def get_tasks_at_level(workflow, level):

    """
    return the tasks at given level

    Args:
        workflow (DAG): DAG object of workflow
        level (int): level which output tasks needed to be

    Returns:
        list(Node)
    """
    tasks = []
    for task in workflow.get_nodes():
        if workflow.get_nodes()[task].get_level() == level:
            tasks.append(workflow.get_nodes()[task])
    return tasks


def get_longest_parent(task):

    """
    Returns the parent of the given task which has the longest execution time

    Args:
        task (Node)

    Returns:
        Node
    """
    longest_parent = task.get_parents()[0]
    for parent in task.get_parents():
        if longest_parent.get_exec_time() < parent.get_exec_time():
            longest_parent = parent
    return longest_parent


def sort_tasks_by_longest_parent(tasks):

    """
    Sort the list of tasks considering the longest parents of the tasks

    Args:
        tasks (list(Node))

    Returns:
        list(Node)
    """
    n = len(tasks)
    if len(tasks) == 1:
        return tasks
    for i in range(n):
        for j in range(0, n - i - 1):
            if get_longest_parent(tasks[j]).get_exec_time() > get_longest_parent(tasks[j + 1]).get_exec_time():
                tasks[j], tasks[j + 1] = tasks[j + 1], tasks[j]
    return tasks


def sort_tasks_by_exec_time(tasks):

    """
    Sort tasks in descending order by the execution time

    Args:
        tasks (list(Node)

    Returns:
        list(Node)
    """
    n = len(tasks)
    for i in range(n):
        for j in range(0, n - i - 1):
            if tasks[j].get_exec_time() < tasks[j+1].get_exec_time():
                tasks[j], tasks[j + 1] = tasks[j + 1], tasks[j]
    return tasks


def get_unassigned_parents(tasks):

    """
    Returns a list of unassigned tasks from a given list of tasks

    Args:
        tasks (list(Node))

    Returns:
        list(Node)
    """
    parents = []
    for parent in tasks.get_parents():
        if not parent.get_is_assigned():
            parents.append(parent)
    return parents


def get_sum_0f_exec_time(cluster):

    """
    Returns the sum of execution time from a given list of tasks

    Args:
        cluster (list(Node))

    Returns:
        int
    """
    exec_sum = 0
    for task in cluster:
        exec_sum += task.get_exec_time()
    return exec_sum


def is_parent_already_assigned(task, parent_id):

    """
    Checks whether the given node is already assigned as parent or not

    Args:
        task (Node)
        parent_id (fw_id)

    Returns:
        int
    """
    parents = task.get_parents()
    for parent in parents:
        if parent.get_fw_id() == parent_id:
            return True
    return False


def is_child_already_assigned(task, child_id):

    """
    Checks whether the given node is already assigned as child or not

    Args:
        task (Node)
        child_id (fw_id)

    Returns:
        int
    """
    children = task.get_children()
    for child in children:
        if child.get_fw_id() == child_id:
            return True
    return False


def assign_parent_to_clusters(task):

    """
    Assign parent tasks of a given node to clusters

    Args:
        task (Node)

    Returns:
        list of clustered Nodes
    """
    cls = []
    c = 0
    # Keep the nodes which are about to cluster
    cluster_c = []
    # Longest parent of the node
    longest_parent = get_longest_parent(task)
    # Level of the nodes
    c_level = longest_parent.get_level()
    max_run_time = 0
    # Check the longest parent is clustered or not. If it is not clustered set the maximum runtime
    if not longest_parent.get_is_assigned():
        c += 1
        cluster_c.append(longest_parent)
        max_run_time = longest_parent.get_exec_time()
        longest_parent._is_assigned = True
    # Get the unassigned parents of node
    par_list = get_unassigned_parents(task)
    if len(par_list) > 0:
        # Sort the unassigned parent tasks in descending order
        par_list = sort_tasks_by_exec_time(par_list)
        c += 1
        cluster_c = []
        # keep the information of the cluster
        cls_info = {}
        # iterate all the unassigned tasks to cluster
        while len(par_list) > 0:
            task = par_list.pop()
            # Check the sum of execution time current cluster and current node is less than or
            # equal to the maximum run time. # If it gets true assign the current task to the cluster
            if (get_sum_0f_exec_time(cluster_c) + task.get_exec_time() <= max_run_time) and len(cluster_c) < 2:
                cluster_c.append(task)
                cls_info[task.get_fw_id()] = {'exec_time': task.get_exec_time(), 'cores': task.get_num_cores()}
                # Mark the task as assigned
                task.isAssigned = True
            else:
                # If the sum of execution time current cluster and current node is larger than the maximum run time.
                # Stop adding node to the current cluster.
                if (len(cls_info)) > 1:  # check the current cluster has more than one node
                    # Create a DAG Node object embedding the current clustered nodes.
                    # Set the fw_id of the first node as the fw_id of the clustered node and
                    # set the sum of execution times of all the nodes as cluster execution time and
                    # maximum number of cores as the cluster required cores.
                    core_space = 0
                    sequential_ids = []
                    first_cluster_node = cluster_c[0]
                    second_cluster_node = cluster_c[1]
                    # Set the task which has maximum number of nodes to the begging of the sequential list
                    if first_cluster_node.get_num_cores() >= second_cluster_node.get_num_cores():
                        core_space = first_cluster_node.get_num_cores()-second_cluster_node.get_num_cores()
                        sequential_ids = [first_cluster_node.get_fw_id(), second_cluster_node.get_fw_id()]
                    elif first_cluster_node.get_num_cores() < second_cluster_node.get_num_cores():
                        core_space = second_cluster_node.get_num_cores() - first_cluster_node.get_num_cores()
                        sequential_ids = [second_cluster_node.get_fw_id(), first_cluster_node.get_fw_id()]

                    cluster = Node(fw_id=sequential_ids[0], level=c_level,
                                   fw_info={'exec_time': get_sum_0f_exec_time(cluster_c),
                                            'cores': cls_info[sequential_ids[0]]['cores']},
                                   parents=[], children=[], assigned=True)
                    cluster.set_cluster_info(cls_info)
                    cluster.set_sequential_ids(sequential_ids)
                    # Set the information about the cluster space available in the cluster to fit other nodes
                    # to reduce the resource utilization.
                    cluster.set_cluster_space([cls_info[cluster.get_fw_ids_to_cluster_sequentially()[1]]['exec_time'], core_space])
                    # Add the children to new clustered node
                    for tsk in cluster_c:
                        children = tsk.get_children()
                        if children is None:
                            continue
                        for child in children:
                            if not is_child_already_assigned(cluster, child.get_fw_id()):
                                cluster.add_child(child)
                    # Add the parents to the new clustered node
                    for tsk in cluster_c:
                        parents = tsk.get_parents()
                        if parents is None:
                            continue
                        for parent in parents:
                            if not is_parent_already_assigned(cluster, parent.get_fw_id()):
                                cluster.add_parent(parent)
                    # Add clustered node as the parent of the children of the clustered tasks
                    for tsk in cluster_c:
                        children = tsk.get_children()
                        if children is None:
                            continue
                        for child in children:
                            child.remove_parent(tsk.get_fw_id())
                            if not is_parent_already_assigned(child, cluster.get_fw_id()):
                                child.add_parent(cluster)
                    # Add clustered node as the children of the parents of the clustered tasks
                    for tsk in cluster_c:
                        parents = tsk.get_parents()
                        if parents is None:
                            continue
                        for parent in parents:
                            parent.remove_child(tsk.get_fw_id())
                            if not is_child_already_assigned(parent, cluster.get_fw_id()):
                                parent.add_child(cluster)
                    cls.append(cluster)
                    # Sets a new cluster
                    cluster_c = []
                    cls_info = {}
                    c += 1
                    cluster_c.append(task)
                    cls_info[task.get_fw_id()] = {'exec_time': task.get_exec_time(), 'cores': task.get_num_cores()}
                    task.isAssigned = True
            # Handle the last task of the list
            if len(par_list) == 0:
                if (len(cls_info)) > 1:
                    core_space = 0
                    sequential_ids = []
                    first_cluster_node = cluster_c[0]
                    second_cluster_node = cluster_c[1]
                    # Set the task which has maximum number of nodes to the begging of the sequential list
                    if first_cluster_node.get_num_cores() >= second_cluster_node.get_num_cores():
                        core_space = first_cluster_node.get_num_cores() - second_cluster_node.get_num_cores()
                        sequential_ids = [first_cluster_node.get_fw_id(), second_cluster_node.get_fw_id()]
                    elif first_cluster_node.get_num_cores() < second_cluster_node.get_num_cores():
                        core_space = second_cluster_node.get_num_cores() - first_cluster_node.get_num_cores()
                        sequential_ids = [second_cluster_node.get_fw_id(), first_cluster_node.get_fw_id()]

                    cluster = Node(fw_id=sequential_ids[0], level=c_level,
                                   fw_info={'exec_time': get_sum_0f_exec_time(cluster_c),
                                            'cores': cls_info[sequential_ids[0]]['cores']},
                                   parents=[], children=[], assigned=True)
                    cluster.set_cluster_info(cls_info)
                    cluster.set_sequential_ids(sequential_ids)
                    # Set the information about the cluster space available in the cluster to fit other nodes to
                    # reduce the resource utilization.
                    cluster.set_cluster_space(
                        [cls_info[cluster.get_fw_ids_to_cluster_sequentially()[1]]['exec_time'], core_space])
                    # Add the children to new clustered node
                    for tsk in cluster_c:
                        children = tsk.get_children()
                        if children is None:
                            continue
                        for child in children:
                            if not is_child_already_assigned(cluster, child.get_fw_id()):
                                cluster.add_child(child)
                    for tsk in cluster_c:
                        parents = tsk.get_parents()
                        if parents is None:
                            continue
                        for parent in parents:
                            if not is_parent_already_assigned(cluster, parent.get_fw_id()):
                                cluster.add_parent(parent)
                    for tsk in cluster_c:
                        children = tsk.get_children()
                        if children is None:
                            continue
                        for child in children:
                            child.remove_parent(tsk.get_fw_id())
                            if not is_parent_already_assigned(child, cluster.get_fw_id()):
                                child.add_parent(cluster)
                    for tsk in cluster_c:
                        parents = tsk.get_parents()
                        if parents is None:
                            continue
                        for parent in parents:
                            parent.remove_child(tsk.get_fw_id())
                            if not is_child_already_assigned(parent, cluster.get_fw_id()):
                                parent.add_child(cluster)
                    cls.append(cluster)
                break
    return cls


def resource_balance(clusters_at_level, tasks, wf):

    """
    Perform resource balancing based on available resource on clustered nodes and update the workflow

    Args:
        wf (DAG)
        clusters_at_level (list(Node): Clustered nodes at a given level
        tasks (list(Node)): Nodes at a given level

    """
    # Iterate the task in level with the all the clustered nodes in the level and
    # check whether is there any available space to fit in the clustered nodes
    for task in tasks:
        parents = task.get_parents()
        for parent in parents:
            # If the parent task is already clustered. Skip
            if len(parent.get_cluster_info()) > 1:
                continue
            # If the clustered task is already filled. Skip
            for cluster in clusters_at_level:
                if bool(cluster.get_fw_ids_to_cluster_parallely()):
                    continue
                # Check the clustered node has space to fit a unclustered task.
                # Task should have less or equal number of core requirement and execution time.
                if cluster.get_cluster_space()[0] >= parent.get_exec_time() and cluster.get_cluster_space()[1] >= parent.get_num_cores():
                    p_parents = parent.get_parents()
                    p_children = parent.get_children()
                    for p_parent in p_parents:
                        # remove the node as child from node's parent node
                        p_parent.remove_child(parent.get_fw_id())
                        # remove the parents from node
                        parent.remove_parent(p_parent.get_fw_id())
                        # Assign the parents of the nodes to the new clustered nodes
                        if not is_parent_already_assigned(cluster, p_parent.get_fw_id()):
                            cluster.add_parent(p_parent)
                        # Assign new cluster as the child of parent nodes of the task
                        if not is_child_already_assigned(p_parent, cluster.get_fw_id()):
                            p_parent.add_child(cluster)
                    for p_child in p_children:
                        # remove the node as parent from node's child node
                        p_child.remove_parent(parent.get_fw_id())
                        # remove the children from node
                        parent.remove_child(p_child.get_fw_id())
                        # Assign the children of the nodes to the new clustered nodes
                        if not is_child_already_assigned(cluster, p_child.get_fw_id()):
                            cluster.add_child(p_child)
                        # Assign new cluster as the parent of child nodes of the task
                        if not is_parent_already_assigned(p_child, cluster.get_fw_id()):
                            p_child.add_parent(cluster)
                    # Delete the node from the WK
                    wf.delete_node(parent.get_fw_id())
                    # Assign a minus key to the refer the parallel running jobs
                    key = -1*randrange(1, 100)
                    # Set parallel nodes to the cluster
                    cluster.set_parallel_ids([cluster.get_fw_ids_to_cluster_sequentially()[1],parent.get_fw_id()], key)
                    # Set sequentially running nodes to the cluster
                    cluster.set_sequential_ids([cluster.get_fw_ids_to_cluster_sequentially()[0], key])


def wpa_clustering(workflow):

    """
    WPA clustering alogirthm

    Args:
        workflow (DAG)

    Returns:
        Clustered workflow DAG (DAG)
    """
    # Iterate the WF level by level
    for level in range(workflow.get_height(), 1, -1):
        cls_at_level = []
        # Get the tasks at level
        tasks_at_level = get_tasks_at_level(workflow, level)
        # Sort the tasks by longest parent of the tasks
        tasks_at_level_sorted = sort_tasks_by_longest_parent(tasks_at_level)
        # Iterate the tasks of the level
        for task in tasks_at_level_sorted:
            # Assign parents of the task to the clusters
            cls = assign_parent_to_clusters(task)
            for cluster in cls:
                cls_info = cluster.get_cluster_info()
                temp_wfid = cluster.get_fw_id()
                for wfid in cls_info:
                    # Delete the tasks which are clustered from the WF
                    workflow.delete_node(wfid)
                # Add new clustered node to the WF
                workflow.add_node(temp_wfid, cluster)
            for cl in cls:
                cls_at_level.append(cl)
        # Resource balance
        resource_balance(cls_at_level, tasks_at_level_sorted, workflow)
    workflow.update_links()
    return workflow


def cluster_vertically(workflow):
    for level in range(1,workflow.get_height()):
        tsk_at_level = get_tasks_at_level(workflow, level)
        for task in tsk_at_level:
            m_task = task
            cluster_c = []
            cluster_c.append(task)
            cls_info = {}
            cls_info[task.get_fw_id()] = {'exec_time': task.get_exec_time(), 'cores': task.get_num_cores()}
            while(True):
                if(task.get_children() is None):
                    break
                if(len(task.get_children())==1):
                    child = task.get_children()[0]
                    if(len(child.get_parents())==1):
                        cls_info[child.get_fw_id()] = {'exec_time': child.get_exec_time(), 'cores': child.get_num_cores()}
                        cluster_c.append(child)
                        task = child
                    else:
                        break
                else:
                    break
            if(len(cls_info)>1):
                # print(cls_info)
                sequential_id = []
                cluster = Node(fw_id=m_task.get_fw_id(), level=level,
                               fw_info={'exec_time': get_sum_0f_exec_time(cluster_c),
                                        'cores': cls_info[task.get_fw_id()]['cores']},
                               parents=[], children=[], assigned=False)
                cluster.set_cluster_info(cls_info)
                for cls in cluster_c:
                    sequential_id.append(cls.get_fw_id())
                cluster.set_sequential_ids(sequential_id)
                if not m_task.get_parents() is None:
                    for parent in m_task.get_parents():
                        cluster.add_parent(parent)
                if not cluster_c[-1].get_children() is None:
                    for child in cluster_c[-1].get_children():
                        child.remove_parent(cluster_c[-1].get_fw_id())
                        child.add_parent(cluster)
                        cluster.add_child(child)
                for tsk in cluster_c:
                    workflow.delete_node(tsk.get_fw_id())
                workflow.add_node(m_task.get_fw_id(), cluster)
                workflow.update_links()
    return workflow




from fireworks.core.rocket_launcher import rapidfire
from fireworks import Firework, FWorker, ScriptTask

from swarmform import SwarmPad
from swarmform.user_objects.firetasks.parallel_tasks import ParallelTask


def main():
	# set up the LaunchPad and reset it
	swarmpad = SwarmPad()
	swarmpad.reset('', require_password=False)

	firetask1 = ScriptTask.from_str('echo "starting"; sleep 30; echo "ending"')
	firetask2 = ScriptTask.from_str('echo "hello from BACKGROUND thread #1"')
	firetask3 = ScriptTask.from_str('echo "hello from BACKGROUND thread #2"')

	# Combine [firetasks] parallely to form a ParallelTask
	par_task = ParallelTask.from_firetasks([firetask1, firetask2])

	# Combine the par_task and firetask3 sequentially
	firework = Firework([par_task, firetask3])

	# store workflow and launch it locally
	swarmpad.add_wf(firework)
	rapidfire(swarmpad, FWorker())


if __name__ == "__main__":
	main()



import datetime

from fireworks import LaunchPad, Firework
from fireworks.fw_config import GRIDFS_FALLBACK_COLLECTION

from swarmform.core.swarmwork import SwarmFlow


class SwarmPad(LaunchPad):

	@classmethod
	def from_dict(cls, d):
		port = d.get('port', None)
		name = d.get('name', None)
		username = d.get('username', None)
		password = d.get('password', None)
		logdir = d.get('logdir', None)
		strm_lvl = d.get('strm_lvl', None)
		user_indices = d.get('user_indices', [])
		wf_user_indices = d.get('wf_user_indices', [])
		ssl = d.get('ssl', False)
		ssl_ca_certs = d.get('ssl_ca_certs',
							 d.get('ssl_ca_file',
								   None))  # ssl_ca_file was the old notation for FWS < 1.5.5
		ssl_certfile = d.get('ssl_certfile', None)
		ssl_keyfile = d.get('ssl_keyfile', None)
		ssl_pem_passphrase = d.get('ssl_pem_passphrase', None)
		authsource = d.get('authsource', None)
		uri_mode = d.get('uri_mode', False)
		mongoclient_kwargs = d.get('mongoclient_kwargs', None)
		return SwarmPad(d['host'], port, name, username, password,
						logdir, strm_lvl, user_indices, wf_user_indices, ssl,
						ssl_ca_certs, ssl_certfile, ssl_keyfile,
						ssl_pem_passphrase,
						authsource, uri_mode, mongoclient_kwargs)

	def reset(self, password, require_password=True, max_reset_wo_password=25):
		"""
		Create a new SwarmForm database. This will overwrite the existing SwarmForm database! To
		safeguard against accidentally erasing an existing database, a password must be entered.
		Args:
			password (str): A String representing today's date, e.g. '2012-12-31'
			require_password (bool): Whether a password is required to reset the DB. Setting to
				false is dangerous because running code unintentionally could clear your DB - use
				max_reset_wo_password to minimize risk.
			max_reset_wo_password (int): A failsafe; when require_password is set to False,
				FWS will not clear DBs that contain more SwarmFlows than this parameter
		"""
		m_password = datetime.datetime.now().strftime('%Y-%m-%d')

		if password == m_password or (
				not require_password and self.workflows.count() <= max_reset_wo_password):
			self.fireworks.delete_many({})
			self.launches.delete_many({})
			self.workflows.delete_many({})
			self.offline_runs.delete_many({})
			self._restart_ids(1, 1, 1)
			if self.gridfs_fallback is not None:
				self.db.drop_collection(
					"{}.chunks".format(GRIDFS_FALLBACK_COLLECTION))
				self.db.drop_collection(
					"{}.files".format(GRIDFS_FALLBACK_COLLECTION))
			self.tuneup()
			self.m_logger.info('SwarmPad was RESET.')
		elif not require_password:
			raise ValueError(
				"Password check cannot be overridden since the size of DB ({} workflows) "
				"is greater than the max_reset_wo_password parameter ({}).".format(
					self.fireworks.count(),
					max_reset_wo_password))
		else:
			raise ValueError(
				"Invalid password! Password is today's date: {}".format(
					m_password))

	def add_sf(self, sf, reassign_all=True):
		"""
		Add SwarmFlow(or firework) to the SwarmPad. The firework ids will be reassigned.
		Args:
			sf (SwarmFlow/Firework)
			reassign_all(bool): Reassign Firework ids
		Returns:
			dict: mapping between old and new Firework ids
		"""

		# Set a new workflow Id, if an id is not provided
		if not hasattr(sf, 'sf_id'):
			sf.sf_id = self.get_new_sf_id()
			self.m_logger.info('Create a new workflow with workflow_id: {}'.format(sf.sf_id))
		else:
			self.m_logger.info('Created a new workflow with workflow_id: {}'.format(sf.sf_id))

		# Check whether the passed workflow is an individual firework
		if isinstance(sf, Firework):
			sf = SwarmFlow.from_Firework(fw=sf, sf_id=sf.sf_id)

		# sets the root FWs as READY
		# prefer to wf.refresh() for speed reasons w/many root FWs
		for fw_id in sf.root_fw_ids:
			sf.id_fw[fw_id].state = 'READY'
			sf.fw_states[fw_id] = 'READY'
		# insert the FireWorks and get back mapping of old to new ids
		old_new = self._upsert_fws(list(sf.id_fw.values()),
								   reassign_all=reassign_all)
		# update the Workflow with the new ids
		sf._reassign_ids(old_new)
		# insert the WFLinks
		self.workflows.insert_one(sf.to_db_dict())
		self.m_logger.info('Added a workflow. id_map: {}'.format(old_new))
		return old_new

	def get_new_sf_id(self, quantity=1):
		"""
		Checkout the next SwarmFlow id
		Args:
			quantity (int): optionally ask for many ids, otherwise defaults to 1
							this then returns the *first* sf_id in that range
		"""
		try:
			return self.fw_id_assigner.find_one_and_update({}, {
				'$inc': {'next_sf_id': quantity}})['next_sf_id']
		except Exception:
			raise ValueError(
				"Could not get next SwarmFlow id! If you have not yet initialized the database,"
				" please do so by performing a database reset (e.g., swarmpad reset)")

	def _restart_ids(self, next_fw_id, next_launch_id, next_sf_id):
		"""
		internal method used to reset firework id counters.
		Args:
			next_fw_id (int): id to give next Firework
			next_launch_id (int): id to give next Launch
			next_sf_id (int): id to give next SwarmFlow
		"""
		self.fw_id_assigner.delete_many({})
		self.fw_id_assigner.find_one_and_replace({'_id': -1},
												 {'next_fw_id': next_fw_id,
												  'next_launch_id': next_launch_id,
												  'next_sf_id': next_sf_id},
												 upsert=True)
		self.m_logger.debug(
			'RESTARTED fw_id, launch_id to ({}, {}, {})'.format(next_fw_id,
																next_launch_id, next_sf_id))

	def get_sf_by_id(self, sf_id):
		"""
		Given a SwarmFlow id, give back the SwarmFlow.
		Args:
			sf_id (int)
		Returns:
			A SwarmFlow object
		"""
		links_dict = self.workflows.find_one({'sf_id': sf_id})

		if not links_dict:
			raise ValueError(
				"Could not find a Workflow with sf_id: {}".format(sf_id))

		fws = map(self.get_fw_by_id, links_dict["nodes"])
		return SwarmFlow(fws, links_dict['links'], links_dict['name'],
						 links_dict['metadata'], links_dict['created_on'],
						 links_dict['updated_on'], None, links_dict['sf_id'])

	def get_sf_by_name(self, sf_name):
		"""
		Given a SwarmFlow name, give back the SwarmFlow.
		If multiple SwarmFlows with the same name are found, first SwarmFlow is returned
		Args:
			sf_name (string)
		Returns:
			A SwarmFlow Object
		"""

		links_dict = self.workflows.find_one({'name': sf_name})

		if not links_dict:
			raise ValueError(
				"Could not find a SwarmFlow with sf_name: {}".format(sf_name))

		fws = map(self.get_fw_by_id, links_dict["nodes"])
		return SwarmFlow(fws, links_dict['links'], links_dict['name'],
						 links_dict['metadata'], links_dict['created_on'],
						 links_dict['updated_on'], None, links_dict['sf_id'])



#!/usr/bin/env python

from setuptools import setup, find_packages
import os

module_dir = os.path.dirname(os.path.abspath(__file__))

if __name__ == "__main__":
    setup(
        name='swarmform',
        version="0.0.1.dev2",
        description='SwarmForm Workflow Management Software',
        long_description=open(os.path.join(module_dir, 'README.md')).read(),
        url='https://github.com/SwarmForm/SwarmForm',
        author='Kalana Wijethunga, Randika Jayasekara, Ayesh Weerasinghe',
        author_email='kalana.16@cse.mrt.ac.lk, rpjayaseka.16@cse.mrt.ac.lk, ayeshweerasinghe.16@cse.mrt.ac.lk',
        packages=find_packages(),
        install_requires=['FireWorks>=1.9.5', 'PyYAML>=5.3.1'],
	classifiers=[
	    "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
	],
        entry_points={
            'console_scripts': [
                'sform = swarmform.scripts.sform_run:sform'
            ]
        }
    )



import datetime
import os
import sys
import signal
import traceback
import six
from argparse import ArgumentParser

from fireworks.scripts.lpad_run import init_yaml, get_output_func

from swarmform import SwarmPad
from swarmform.core.swarmwork import SwarmFlow
from swarmform.core.cluster import cluster_sf
from swarmform.sf_config import LAUNCHPAD_LOC, CONFIG_FILE_DIR

DEFAULT_LPAD_YAML = "my_swarmpad.yaml"


def handle_interrupt(signum):
    sys.stderr.write("Interrupted by signal {:d}\n".format(signum))
    sys.exit(1)


def get_sp(args):
    try:
        if not args.launchpad_file:
            if os.path.exists(os.path.join(args.config_dir, DEFAULT_LPAD_YAML)):
                args.launchpad_file = os.path.join(args.config_dir, DEFAULT_LPAD_YAML)
            else:
                args.launchpad_file = LAUNCHPAD_LOC

        if args.launchpad_file:
            return SwarmPad.from_file(args.launchpad_file)
        else:
            args.loglvl = 'CRITICAL' if args.silencer else args.loglvl
            return SwarmPad(logdir=args.logdir, strm_lvl=args.loglvl)
    except Exception:
        traceback.print_exc()
        err_message = 'SwarmForm was not able to connect to MongoDB. Is the server running? ' \
                      'The database file specified was {}.'.format(args.launchpad_file)
        if not args.launchpad_file:
            err_message += ' Type "sform init" if you would like to set up a file that specifies ' \
                           'location and credentials of your Mongo database (otherwise use default ' \
                           'localhost configuration).'
        raise ValueError(err_message)


def reset(args):
    sp = get_sp(args)
    if not args.password:
        if input('Are you sure? This will RESET {} workflows and all data. (Y/N)'.format(
                sp.workflows.count()))[0].upper() == 'Y':
            args.password = datetime.datetime.now().strftime('%Y-%m-%d')
        else:
            raise ValueError('Operation aborted by user.')
    sp.reset(args.password)


def add_sf(args):
    sp = get_sp(args)
    if args.dir:
        files = []
        for f in args.sf_file:
            files.extend([os.path.join(f, i) for i in os.listdir(f)])
    else:
        files = args.sf_file
    for f in files:
        fwf = SwarmFlow.from_file(f)
        if args.check:
            from fireworks.utilities.dagflow import DAGFlow
            DAGFlow.from_fireworks(fwf)
        sp.add_sf(fwf)


def get_sf(args):
    if args.sf_id:
        sf = get_sf_by_id(args)
    if args.name:
        sf = get_sf_by_name(args)
    # TODO: Use get_wf_summary_dict function and print formatted output
    print(args.output(sf))


def get_sf_by_id(args):
    sp = get_sp(args)

    if isinstance(args.sf_id, int):
        swarmflow = sp.get_sf_by_id(args.sf_id)
        return swarmflow
    else:
        raise ValueError('Incorrect id type {}.'.format(type(args.sf_id)))


def get_sf_by_name(args):
    sp = get_sp(args)
    swarmflow = sp.get_sf_by_name(args.sf_name)
    return swarmflow


# Cluster the jobs in a workflow
def cluster_workflow(args):
    sp = get_sp(args)
    unclustered_sf = sp.get_sf_by_id(args.sf_id)
    unclustered_sf_fw_id = unclustered_sf.fws[0].fw_id
    clustered_workflow = cluster_sf(sp, args.sf_id)
    sp.add_sf(clustered_workflow)
    sp.archive_wf(unclustered_sf_fw_id)
    sp.m_logger.info('Workflow with id {} clustered succesfully'.format(args.sf_id))


def sform():
    m_description = 'A command line interface to SwarmForm. For more help on a specific command, ' \
                    'type "sform <command> -h".'

    parser = ArgumentParser(description=m_description)
    parser.add_argument('-o', '--output', choices=["json", "yaml"],
                        default='json', type=lambda s: s.lower(),
                        help='Set output display format to either json or YAML. '
                             'YAML is easier to read for long documents. JSON is the default.')

    parser.add_argument('-l', '--launchpad_file', help='path to LaunchPad file containing '
                                                       'central DB connection info',
                        default=None)
    parser.add_argument('-c', '--config_dir',
                        help='path to a directory containing the LaunchPad file (used if -l unspecified)',
                        default=CONFIG_FILE_DIR)
    parser.add_argument('--logdir', help='path to a directory for logging')
    parser.add_argument('--loglvl', help='level to print log messages', default='INFO')
    parser.add_argument('-s', '--silencer', help='shortcut to mute log messages', action='store_true')

    subparsers = parser.add_subparsers(help='command', dest='command')

    init_parser = subparsers.add_parser(
        'init', help='Initialize a SwarmForm launchpad YAML file.')
    init_parser.add_argument('-u', '--uri_mode',
                             action="store_true",
                             help="Connect via a URI, see: "
                                  "https://docs.mongodb.com/manual/reference/connection-string/")
    init_parser.add_argument('--config-file', default=DEFAULT_LPAD_YAML,
                             type=str,
                             help="Filename to write to.")
    init_parser.set_defaults(func=init_yaml)

    reset_parser = subparsers.add_parser('reset', help='Reset and re-initialize the SwarmForm database')
    reset_parser.add_argument('--password', help="Today's date,  e.g. 2012-02-25. "
                                                 "Password or positive response to input prompt "
                                                 "required to protect against accidental reset.")
    reset_parser.set_defaults(func=reset)

    addsf_parser = subparsers.add_parser('add', help='Insert a SwarmFlow from file')
    addsf_parser.add_argument('-sf', '--sf_file', nargs='+',
                              help='Path to a Firework or SwarmFlow file')
    addsf_parser.add_argument('-d', '--dir',
                              action="store_true",
                              help="Directory mode. Finds all files in the "
                                   "paths given by wf_file.")
    addsf_parser.set_defaults(func=add_sf, check=False)

    getsf_parser = subparsers.add_parser('get_sf', help='Get SwarmFlow from SwarmPad')
    getsf_parser.add_argument('-id', '--sf-id', type=int, help='SwarmFlow id')
    getsf_parser.add_argument('-n', '--name', type=str, help='SwarmFlow name')
    getsf_parser.set_defaults(func=get_sf)

    cluster_wf_parser = subparsers.add_parser('cluster',
                                              help='Cluster the fireworks in the SwarmFlow and save the new '
                                                   'SwarmFlow to the database')
    cluster_wf_parser.add_argument('-sf', '--sf_id', help='Id of the SwarmFlow to cluster', default=None, type=int)
    cluster_wf_parser.set_defaults(func=cluster_workflow)

    args = parser.parse_args()

    args.output = get_output_func(args.output)

    if args.command is None:
        # if no command supplied, print help
        parser.print_help()
    else:
        if hasattr(args, "fw_id") and args.fw_id is not None and \
                isinstance(args.fw_id, six.string_types):
            if "," in args.fw_id:
                args.fw_id = [int(x) for x in args.fw_id.split(",")]
            else:
                args.fw_id = [int(args.fw_id)]

        args.func(args)

    signal.signal(signal.SIGINT, handle_interrupt)  # graceful exit on ^C


if __name__ == '__main__':
    sform()
